#
#  engine
#
#  Playout Daemon for autoradio project
#
#
#  Copyright (C) 2017-2018 Gottfried Gaisbauer <gottfried.gaisbauer@servus.at>
#
#  This file is part of engine.
#
#  engine is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  engine is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with engine. If not, see <http://www.gnu.org/licenses/>.
#

import sys
import time
import redis
import logging
import threading

from datetime import datetime
from threading import Event

from modules.communication.redis.messenger import RedisMessenger
from modules.communication.redis.statestore import RedisStateStore
# from modules.communication.connection_tester import ConnectionTester
from modules.base.exceptions import RedisConnectionException
from modules.base.enum import RedisChannel, FallbackType
from modules.base.utils import TerminalColors


# ------------------------------------------------------------------------------------------ #
class ServerRedisAdapter(threading.Thread, RedisMessenger):
    debug = False
    pubsub = None
    config = None
    redisdb = None
    channel = ""
    scheduler = None
    redisclient = None
#    connection_tester = None
    soundsystem = None
    socket = None
    
    # ------------------------------------------------------------------------------------------ #
    def __init__(self, config):
        threading.Thread.__init__(self)
        RedisMessenger.__init__(self, config)

        # init
        #threading.Thread.__init__ (self)
        self.config = config
        self.shutdown_event = Event()

        self.channel = RedisChannel.STANDARD.value
        self.section = ''
        self.rstore = RedisStateStore(config)
        self.errnr = '00'
        self.components = {'controller':'01', 'scheduling':'02', 'playd':'03', 'recorder':'04', 'helpers':'09'}
        self.fromMail = ''
        self.adminMails = ''
        self.can_send = None
        self.redisclient = ClientRedisAdapter(config)
#        self.connection_tester = ConnectionTester()

    # ------------------------------------------------------------------------------------------ #
    def run(self):
        self.redisdb = redis.Redis(host=self.config.get("redis_host"), port=self.config.get("redis_port"), db=self.config.get("redis_db"))
        self.pubsub = self.redisdb.pubsub()
        self.pubsub.subscribe(self.channel)

        self.logger.debug(TerminalColors.YELLOW.value + "waiting for REDIS message on channel " + self.channel + TerminalColors.ENDC.value)

        # listener loop
        for item in self.pubsub.listen():
            if item["type"] == "subscribe":
                continue

            self.logger.debug(TerminalColors.YELLOW.value + "received REDIS message: " + TerminalColors.ENDC.value + str(item))

            item["channel"] = self.decode_if_needed(item["channel"])
            item["data"] = self.decode_if_needed(item["data"])

            try:
                self.work(item)
            except RedisConnectionException as rce:
                self.logger.error(str(rce))

            if not self.shutdown_event.is_set():
                self.logger.debug(TerminalColors.YELLOW.value + "waiting for REDIS message on channel " + self.channel + TerminalColors.ENDC.value)

        self.pubsub.unsubscribe()

        if not self.shutdown_event.is_set():
            self.logger.warning("unsubscribed from " + self.channel + " and finished")

    # ------------------------------------------------------------------------------------------ #
    def decode_if_needed(self, val):
        if isinstance(val, bytes):
            return val.decode("utf-8")
        return val

    # ------------------------------------------------------------------------------------------ #
    def listen_for_one_message(self, channel, socket_timeout=2):
        self.redisdb = redis.Redis(host=self.config.get("redis_host"), port=self.config.get("redis_port"), db=self.config.get("redis_db"), socket_timeout=socket_timeout)
        self.pubsub = self.redisdb.pubsub()
        self.pubsub.subscribe(channel)

        try:
            self.logger.debug("I am listening on channel '"+channel+"' for "+str(socket_timeout)+" seconds")

            for item in self.pubsub.listen():
                it = self.receive_message(item)
                if it is not None:
                    break
        except redis.exceptions.TimeoutError as te:
            raise te

        return item["data"]

    # ------------------------------------------------------------------------------------------ #
    def receive_message(self, item):
        if item["type"] == "subscribe":
            self.logger.info("i am subscribed to channel " + item["channel"].decode("utf-8"))
            return None

        item["channel"] = item["channel"].decode("utf-8")
        if isinstance(item["data"], bytes):
            item["data"] = item["data"].decode("utf-8")

        self.pubsub.unsubscribe()
        return item

    # ------------------------------------------------------------------------------------------ #
    def work(self, item):
        if item["data"] == "fetch_new_programme":
            #self.execute(RedisChannel.FNP_REPLY.value, self.scheduler.fetch_new_programme)
            self.execute(RedisChannel.FNP_REPLY.value, self.scheduler.get_act_programme_as_string)

        elif item["data"] == "shutdown":
            self.terminate()

        elif item["data"] == "init_player":
            self.execute(RedisChannel.IP_REPLY.value, self.soundsystem.init_player)

        elif item["data"] == "get_act_programme":
            self.execute(RedisChannel.GAP_REPLY.value, self.scheduler.get_act_programme_as_string)

        elif item["data"] == "get_status":
            def get_status_string():
                status = self.soundsystem.monitoring.get_status()
                return str(status)

            self.execute(RedisChannel.GS_REPLY.value, get_status_string)

#        elif item["data"] == "get_connection_status":
#            self.execute(RedisChannel.GCS_REPLY.value, self.connection_tester.get_connection_status)

        elif item["data"] == "print_message_queue":
            self.execute(RedisChannel.PMQ_REPLY.value, self.scheduler.print_message_queue)

        elif item["data"].find("set_next_file") >= 0:
            playlist = item["data"].split()[1]
            playlist = playlist[0:len(playlist)-8]
            self.execute(RedisChannel.SNF_REPLY.value, self.scheduler.set_next_file_for, playlist)

        elif item["data"].find("get_next_file") >= 0:
            playlist = item["data"].split()[1]
            #playlist = playlist[0:len(playlist)-8]
            self.execute(RedisChannel.GNF_REPLY.value, self.scheduler.get_next_file_for, playlist)

        elif item["data"].find("on_play") >= 0:
            source = item["data"].split("on_play ")[1]
            self.execute(RedisChannel.TS_REPLY.value, self.scheduler.soundsystem.on_play, source)

        elif item["data"] == "recreate_db":
            self.execute(RedisChannel.RDB_REPLY.value, self.scheduler.recreate_database)


        else:
            raise RedisConnectionException("ServerRedisAdapter Cannot understand command: " + item["data"])

    # ------------------------------------------------------------------------------------------ #
    def execute(self, channel, f, param1=None, param2=None, param3=None):
        if param1 != None:
            if param2 != None:
                if param3 != None:
                    reply = f(param1, param2, param3)
                else:
                    reply = f(param1, param2)
            else:
                reply = f(param1)
        else:
            reply = f()

        if reply is None:
            reply = ""

        # sometimes the sender is faster than the receiver. redis messages would be lost
        time.sleep(0.1)

        self.logger.debug(TerminalColors.YELLOW.value + "replying REDIS message " + TerminalColors.ENDC.value + reply + TerminalColors.YELLOW.value + " on channel " + channel + TerminalColors.ENDC.value)

        # publish
        self.redisclient.publish(channel, reply)

    # ------------------------------------------------------------------------------------------ #
    def join_comm(self):
        try:
            while self.is_alive():
                self.logger.debug(str(datetime.now())+" joining")
                self.join()
                self.logger.warning("join out")

        except (KeyboardInterrupt, SystemExit):
            # Dem Server den Shutdown event setzen
            # server.shutdown_event.set()
            # Der Server wartet auf Eingabe
            # Daher einen Client initiieren, der eine Nachricht schickt
            self.halt()
            sys.exit('Terminated')

    # ------------------------------------------------------------------------------------------ #
    def halt(self):
        """
        Stop the server
        """
        if self.shutdown_event.is_set():
            return
        self.shutdown_event.set()
        try:
            self.socket.unbind("tcp://"+self.ip+":"+self.port)
        except:
            pass

        # self.socket.close()

    # ------------------------------------------------------------------------------------------ #
    def send(self, message):
        """
        Send a message to the client
        :param message: string
        """

        # FIXME Review logic
        if not self.can_send:
            self.logger.debug("sending a "+str(len(message))+" long message via REDIS.")
            self.socket.send(message.encode("utf-8"))
            self.can_send = False
        else:
            self.logger.warning("cannot send message via REDIS: "+str(message))



    def terminate(self):
        """
        Called when thread is stopped or a signal to terminate is received.
        """
        self.shutdown_event.set()
        self.scheduler.terminate()
        self.pubsub.close()
        self.logger.info("Shutdown event received. Bye bye ...")


# ------------------------------------------------------------------------------------------ #
class ClientRedisAdapter(RedisMessenger):

    def __init__(self, config):
        RedisMessenger.__init__(self, config)

    # ------------------------------------------------------------------------------------------ #
    def publish(self, channel, message):
        if type(channel) == RedisChannel:
            channel = channel.value

        self.rstore.publish(channel, message)
